/****************************************************************************
 * This file is part of the program ALIFRITZ                                *
 *                                                                          *
 *    Copyright (C) 2001  Roland Fleissner                                  *
 *    e-mail: fleissne@uidaho.edu                             		    * 
 *                                                                          *
 *  This program is free software; you can redistribute it and/or modify    *
 *  it under the terms of the GNU General Public License as published by    *
 *  the Free Software Foundation; either version 2 of the License, or       *
 *  (at your option) any later version.                                     *
 *                                                                          *
 *  This program is distributed in the hope that it will be useful,         *
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of          *
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           *
 *  GNU General Public License for more details.                            *
 *                                                                          *
 *  You should have received a copy of the GNU General Public License       *
 *  along with this program; if not, write to the                           *
 *  Free Software Foundation, Inc., 59 Temple Place, Suite 330,             *
 *  Boston, MA  02111-1307  USA                                             *
 ****************************************************************************/

#include <vector>
#include <iostream>
#include <cstdio>
#include <fstream>
#include <cstdlib>
#include <ctime>
#include <string>
#include <list>
#include <climits>
#include <cfloat>

using namespace std;

#include "seqcode.hh"
#include "parser.hh"
#include "eingabe.hh"
#include "data.hh"
#include "indel.hh"
#include "substmod.hh"
#include "treespec.hh"
#include "mult_align.hh"
#include "random.hh"

#define TIME_TO_OPT 100

int main(int argc,char* argv[]) {

    initrandom();
    //zzinitrandom();//falls mit einer bestimmten Zufallszahl aus der Datei 
                     //Zufallszahl initialisiert werden soll: initrandom durch zzinitrandom ersetzen
    
    // Eingabeparameter einlesen
    if(argc!=3){
       cerr << "Please invoke alifritz with the name of the control file and the type of analysis as arguments.\n";
       exit(EXIT_FAILURE);
    }
    eingabe_parameter ep;
    ep.from_file(argv[1]);
    
    int antyp;
    switch(argv[2][0]) {
        case '0' :
	   antyp=0; // ohne alles
	   break;
	case '1' :
           antyp=1; // NJ
	   break;
	case '2' :
	   antyp=2; // NJ und NNI
	   break;
	default :
	   cerr << "unknown option "<<argv[2][0]<<endl;
	   exit(EXIT_FAILURE);
    }

    // Alignment und Baum nach Vorgaben konstruieren
    double likelihood=-DBL_MAX;
    ofstream logfile(ep.fil.logfile.c_str());
    if(!logfile) {
      cerr << "Cannot open logfile "<<ep.fil.logfile.c_str()<<endl;
      exit(EXIT_FAILURE);
    }
    mult_alignment *p=new mult_alignment(ep,antyp), *maxal=new mult_alignment(*p);
    unsigned int unchanged=0;
    unsigned int nnicount=0;
    unsigned int nnimaxl=0;
    while(p->inrange() && nnicount<20*p->spez()+1) {
	if(unchanged%(TIME_TO_OPT*p->spez())==0 && unchanged>0 && !p->burnin()) {
	   if(nnicount<20*p->spez() && antyp==2)
	      p->nearestneighbor(p->spez());
	   ++nnicount;
	}
        if(likelihood<p->likelihood()-1.0 || likelihood>p->likelihood()+1.0)
	   unchanged=0;
	else
	   ++unchanged;
	 
        if(p->report()) {
	   p->bericht(logfile);
	   cout<<"log likelihood after "<<p->step()<<" steps: "<<p->likelihood()<<endl;
	}
	likelihood = p->likelihood();
        p->maunewton();
	p->sampparams();
	if(p->likelihood()>maxal->likelihood()) {
	   nnicount=0; //pfuipfui
           delete maxal;
	   maxal=new mult_alignment(*p);
	   if(antyp==2)
	      nnimaxl=nnicount;
	}
    }
    p->bericht(logfile);
    logfile<<"Time: "<<clock()/CLOCKS_PER_SEC<<endl;
    logfile.close();
    cout<<"final log likelihood (after "<<p->step()<<" steps): "<<p->likelihood()<<endl;
    cout<<"highest log likelihood: "<<maxal->likelihood()<<endl;

    // Alignment mit der besten Likelihood im Phylip-Format ausgeben
    ofstream outalign(ep.fil.outalign.c_str(),ios::app);
    if(!outalign) {
      cerr << "Cannot open output file "<<ep.fil.outalign.c_str()<<endl;
      exit(EXIT_FAILURE);
    }
    maxal->print_phylip(outalign);
    outalign.close();

    // Baum mit der hoechsten likelihood ausgeben
    ofstream outtree(ep.fil.outtree.c_str(),ios::app);
    if(!outtree) {
      cerr << "Cannot open output file "<<ep.fil.outtree.c_str()<<endl;
      exit(EXIT_FAILURE);
    }
    maxal->print_tree(outtree);
    outtree.close();
    
    // Parameter mit der hoechsten likelihood ausgeben
    ofstream outparam(ep.fil.outparam.c_str(),ios::app);
    if(!outparam) {
      cerr << "Cannot open output file "<<ep.fil.outparam.c_str()<<endl;
      exit(EXIT_FAILURE);
    }
    maxal->print_params(outparam);
    outparam << "likelihood = "<<maxal->likelihood()
             << "\nreached after "<<maxal->step()<<" iterations"
	     << "\nand "<<nnimaxl<<" NNIs\n";
    outparam.close();
    
    return 0;
}

