/****************************************************************************
 * This file is part of the program ALIFRITZ                                *
 *                                                                          *
 *    Copyright (C) 2001  Roland Fleissner                                  *
 *    e-mail: fleissne@uidaho.edu                            		    * 
 *                                                                          *
 *  This program is free software; you can redistribute it and/or modify    *
 *  it under the terms of the GNU General Public License as published by    *
 *  the Free Software Foundation; either version 2 of the License, or       *
 *  (at your option) any later version.                                     *
 *                                                                          *
 *  This program is distributed in the hope that it will be useful,         *
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of          *
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           *
 *  GNU General Public License for more details.                            *
 *                                                                          *
 *  You should have received a copy of the GNU General Public License       *
 *  along with this program; if not, write to the                           *
 *  Free Software Foundation, Inc., 59 Temple Place, Suite 330,             *
 *  Boston, MA  02111-1307  USA                                             *
 ****************************************************************************/

#include <string>
#include <vector>
#include <iostream>
#include <iomanip>
#include <fstream>
#include <cstdio>
#include <cmath>
#include <cstdlib>
#include <climits>  

using namespace std;

#include "seqcode.hh"
#include "data.hh"


//+++++++++++++++++++++++++++++++++++++++++++
//             class dataset                +
//+++++++++++++++++++++++++++++++++++++++++++


dataset::dataset()
{
}

dataset::dataset(ifstream& datei,int seqformat)
{
  switch(seqformat) {
    case 0: //FASTA
      fromfasta(datei);
      break;
    case 1: //PHYLIP
      fromphylip(datei);  
      break;
    default:
      cerr << "Format " << seqformat << " not supported\n";
      exit(EXIT_FAILURE);
  }  
  datei.close();
  
  guesstype();
}

dataset::~dataset() {
}

void dataset::fromfasta(ifstream& datei)
{
      unsigned int i, flag;
      bool flag2;
      char c;

      i = 0;
      flag = 2;
      flag2=TRUE;
      while(datei.get(c))
      {
         switch(c) {
           case '>' :
             flag = 1;
	     flag2=TRUE;
	     i++;
	     Identif.resize(i);
	     Seqs.resize(i);
	     break;
           case '\n' :
	     if(flag==2) {
	        cerr << "The sequence file is not in FASTA format!\n";
		exit(EXIT_FAILURE);
	     }
             flag = 0;
	     break;
           default :
	     if(flag==2) {
	        cerr << "The sequence file is not in FASTA format!\n";
		exit(EXIT_FAILURE);
	     }
             if(isprint(c) && !isspace(c))
	     {
	       if(flag==1 && flag2) {
	           Identif[i-1].push_back(c); // Name nur bis zum ersten Leerzeichen
	       } else if(flag==0)
	       {
	         if(c!='-')
	           Seqs[i-1].push_back(toupper(c));
		 else
		   Seqs[i-1].push_back('_');
	       }
	     } else {
	       if(flag==1 && Identif[i-1].size()>0) // leading blanks erlaubt
	          flag2=FALSE;
             }
         }
      }
      if(Identif.size()==0) {
         cerr << "There is no data!\n";
	 exit(EXIT_FAILURE);
      }
      for(i=0;i<Identif.size();++i)
          if(Identif[i].size()==0) {
	     cerr << "There are sequences without names!\n";
	     exit(EXIT_FAILURE);
	  }
}

void dataset::fromphylip(ifstream& datei)
{
      unsigned int Maxspc, Maxseqc;
      unsigned int i, j, k, leerZ, Zeilen, Idcount, flag;
      vector<unsigned int> Position(1);
      string firstline="";
      char c;


      while(datei.get(c) && c!='\n')
          firstline+=c;
      if(sscanf(firstline.c_str(),"%u %u",&Maxspc,&Maxseqc)!=2) {
         cerr << "The sequence file is not in PHYLIP format\n";
	 exit(EXIT_FAILURE);
      }
      Identif.resize(Maxspc);
      for(i=0;i<Identif.size();i++)
      {
        Identif[i].resize(10);
        for(j=0;j<Identif[i].size();j++)
          Identif[i][j]='\0';
      }
      Seqs.resize(Maxspc);
      for(i=0;i<Seqs.size();i++)
      {
        Seqs[i].resize(Maxseqc);
        for(j=0;j<Seqs[i].size();j++)
          Seqs[i][j]='\0';
      }
      Position.resize(Maxspc);
      for(i=0;i<Position.size();i++)
        Position[i]=0;
      leerZ = 0;
      Zeilen = 0;
      Idcount = 0;
      flag = 0;
  
      do
      {
        if(c=='\n')
        {
          if(flag==0)
            leerZ++;
          Zeilen++;
          Idcount = 0;
          flag = 0;
          if((Zeilen-leerZ)%Seqs.size()==0)
          {
            //Kontrolle ob Zeilen gleich lang
            for(i=0;i<Position.size()-1;i++)
            {
              if(Position[i]!=Position[i+1])
              {
                cerr << "Something's wrong\n"
                     << "Position[" << i << "]="
	             << Position[i] << " != Position["
	             << i+1 << "]=" << Position[i+1] << "\n";
                exit(1);
              }
	    }
          }
        }
        else
        {
          if(isprint(c) && !isspace(c))
          {
            flag = 1;
	    if(Zeilen==0)
	    {
	      cerr << "Please remove text in the first line\n";
	      exit(1);
	    }
          }
          if(flag==1)
          {
            Idcount++;
            if(Idcount<=Identif[0].size() && (Zeilen-leerZ)<Identif.size())
              Identif[Zeilen-leerZ][Idcount-1]=c;
            else
            {
              if(Position[(Zeilen-leerZ)%Seqs.size()]<Seqs[0].size())
	      {
                if(isalpha(c) || c=='-'  || c=='_' || c=='*' || c=='?')
	        {
		  if(c!='-')
	            Seqs[(Zeilen-leerZ)%Seqs.size()][Position[(Zeilen-leerZ)%Seqs.size()]]=toupper(c);
		  else
		    Seqs[(Zeilen-leerZ)%Seqs.size()][Position[(Zeilen-leerZ)%Seqs.size()]]='_';
	          Position[(Zeilen-leerZ)%Seqs.size()]++;
	        }
	        else
	        {
	          if(c=='.')
	          {
	            if((Zeilen-leerZ)%Seqs.size()>0)
	            {
	              Seqs[(Zeilen-leerZ)%Seqs.size()][Position[(Zeilen-leerZ)%Seqs.size()]]=Seqs[0][Position[(Zeilen-leerZ)%Seqs.size()]];
	              Position[(Zeilen-leerZ)%Seqs.size()]++;
	            }
	            else
	            {
	              cerr << "Error in Phylip file: dot in the first sequence\n";
		      exit(1);
	            }
	          }
	        }
	      }
            }
          }
        }
  
      } while(datei.get(c) && Position[Position.size()-1]<Seqs[0].size());
  
      //Kontrolle
      for(i=0;i<Position.size()-1;i++)
      {
        if(Position[i]!=Position[i+1])
        {
          cerr << "Something's wrong\n"
               << "Position[" << i << "]="
	       << Position[i] << " != Position["
	       << i+1 << "]=" << Position[i+1] << "\n";
          exit(1);
        }
        if(Position[i]!=Seqs[0].size() || Position[i+1]!=Seqs[0].size())
        {
          cerr << "Something's wrong.\n"
               << "Not enough nucleotides in the sequence\n"
	       << i+1 << " or " << i+2 << "\n";
          exit(1);      
        }
       }
      //leading blanks in Identif
      for(i=0;i<Identif.size();i++)
      {
        flag = 0;
        for(j=1;j<=Identif[0].size();j++)
        {
          if(isprint(Identif[i][Identif[0].size()-j]) && !isspace(Identif[i][Identif[0].size()-j]))
            flag=1;
          if(isspace(Identif[i][Identif[0].size()-j]) && flag==1)
            Identif[i][Identif[0].size()-j]='_';
        }
      }
      //gleiche Namen
      for(i=0;i<Identif.size();i++)
        for(j=i+1;j<Identif.size();j++)
        {
          flag = 0;
          for(k=0;k<Identif[0].size();k++)
            if(Identif[i][k]!=Identif[j][k])
	      flag = 1;
          if(flag == 0)
          {
            cerr << "Sequences " << i+1 << " and "
	         << j+1 << " have the same name\n";
	    exit(1);
          }
        }
      
}


void dataset::guesstype()
{
     bool Uflag, nonDNAflag;
     unsigned int i, j;
     
     Uflag=nonDNAflag=FALSE;
     for(i=0;i<Seqs.size();i++)
        for(j=0;j<Seqs[i].size();j++) {
	   if(Seqs[i][j]=='U')
	      Uflag=TRUE;
	   else 
	      if(Seqs[i][j]=='E' || Seqs[i][j]=='F' || 
	         Seqs[i][j]=='I' || Seqs[i][j]=='L' ||
	         Seqs[i][j]=='P' || Seqs[i][j]=='Q' ||
	         Seqs[i][j]=='X' || Seqs[i][j]=='Z')
                    nonDNAflag=TRUE;
	}

     if(Uflag && nonDNAflag) {
        cerr << "Don't know what to do: there are U's in\n"
	     << "an amino acid sequence\n";
	exit(EXIT_FAILURE);
     } else {
        if(nonDNAflag)
	  typ=PROTEIN;
	else 
	  typ=DNA;
     }
}


void dataset::print_phylip(ostream& ostr)
{
   unsigned int i, j, Maxspc, Maxseqc;
   
   Maxspc = spez();
   Maxseqc = laen(0);
   
   ostr << setw(10) << Maxspc
        << setw(10) << Maxseqc << endl;
   for(i=0;i<Maxspc;i++)
   {
     for(j=0;j<Maxseqc;j++)
       if(Seqs[i][j]=='_')
          Seqs[i][j]='-';
     for(j=0;j<Identif[i].size() && j<10;j++)
       ostr << Identif[i][j];
     while(j<10) {
       ostr << " ";
       j++;
     }
     ostr << Seqs[i] << endl;	
   }
}

void dataset::print_screen(ostream& ostr)
{
   vector<string> linewise;
   int i;
   unsigned int j, Maxspc;
   
   Maxspc = spez();
   
   linewise.resize(Maxspc);
   for(j=0;j<Maxspc;j++)
     linewise[j]="";
   i = 0;
   do {
     for(j=0;j<Maxspc;j++)
     {
       linewise[j]+=Seqs[j][i]; 
     }
     i++;
     if(i%70==0){
       for(j=0;j<Maxspc;j++)
       {
          ostr << linewise[j] << endl;
	  linewise[j]=""; 
       }
       ostr << endl;
     }
   } while (i!=laen(0));

   if(linewise[0]!="")
   {
     for(j=0;j<Maxspc;j++)
        ostr << linewise[j] << endl;
     ostr << endl;
   }
}

unsigned int dataset::spez() const
{
  return Identif.size();
}

int dataset::laen(int s) const // gibt Laenge der Sequenz s zurueck (angefangen bei 0)
{
  if((unsigned)s>=Seqs.size())
  {
    cerr << "The specified sequence number is too big.\n"
         << "s=" << s
	 << "There are only " << Seqs.size() << " in the data set.\n"
	 << "Their numbering starts with 0.\n";
    exit(1);
  }
  
  return Seqs[s].size();
}

void dataset::getfreqs(vector<double>& f) // ermittelt absolute Zustandshaeufigkeiten
{
   f.resize(typ);
   for(int i=0;i<typ;i++)
      f[i]=0.0;
   switch(typ) {
      case DNA :
         getACGT(f);
         break;
      case PROTEIN :
         getAAfreqs(f);
         break;
      default :
         cerr << "Unknown data type in function getfreqs\n";
	 exit(EXIT_FAILURE);
   }
}

void dataset::getACGT(vector<double>& f) // ermittelt absolute Basenhaeufigkeiten
{
  unsigned int i, j, k;
  vector<int> states;
  
  for(i=0;i<Identif.size();i++)
    for(j=0;j<Seqs[i].size();j++) {
      chartobase(Seqs[i][j],states);
      for(k=0;k<states.size();k++)
         f[states[k]]+=1.0/(double)(states.size());
    }
}

void dataset::getAAfreqs(vector<double>& f) // ermittelt absolute Aminosaeurehaeufigkeiten
{
  unsigned int i, j, k;
  vector<int> states;
  
  for(i=0;i<Identif.size();i++)
    for(j=0;j<Seqs[i].size();j++) {
      chartoamac(Seqs[i][j],states);
      for(k=0;k<states.size();k++)
         f[states[k]]+=1.0/(double)(states.size());
    }
}


ostream& operator<<(ostream& ostr,const vector<char>& d) {
    for (unsigned int i=0; i<d.size(); ++i){
	ostr << d[i];
    }
    return ostr;
}

/***********************************************************************************
 ****************** class bioseq         *******************************************
 ***********************************************************************************/


bioseq::bioseq()
{}

bioseq::bioseq(vector<char>& seq, int typ)
{
   unsigned int i;
   char c;
   vector<int> states;

   for(i=0;i<seq.size();i++) {
      c = seq[i];
      if(c!='_' && c!='-') {
         switch(typ) {
             case DNA :
	       push_back(chartobase(c));
	       break;
             case PROTEIN :
	       push_back(chartoamac(c));
	       break;
	     default :
	       cerr << "Unknown data typ in function bioseq(vector<char>& seq, alphsize typ)\n";
	       exit(EXIT_FAILURE);
         }
      }
   }
}

bioseq::~bioseq()
{}

int bioseq::laen() const
{
    return (int)size();
}

