/****************************************************************************
 * This file is part of the program alifritz                                *
 *                                                                          *
 *    Copyright (C) 2001  Roland Fleissner                                  *
 *    e-mail: fleissne@uidaho.edu                            		    * 
 *                                                                          *
 *  This program is free software; you can redistribute it and/or modify    *
 *  it under the terms of the GNU General Public License as published by    *
 *  the Free Software Foundation; either version 2 of the License, or       *
 *  (at your option) any later version.                                     *
 *                                                                          *
 *  This program is distributed in the hope that it will be useful,         *
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of          *
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           *
 *  GNU General Public License for more details.                            *
 *                                                                          *
 *  You should have received a copy of the GNU General Public License       *
 *  along with this program; if not, write to the                           *
 *  Free Software Foundation, Inc., 59 Temple Place, Suite 330,             *
 *  Boston, MA  02111-1307  USA                                             *
 ****************************************************************************/

#include <iostream>
#include <fstream>
#include <string>
#include <vector>
#include <list>
#include <cmath>
#include <cfloat>
#include <cstdio>

using namespace std;

#include "seqcode.hh"
#include "parser.hh"
#include "eingabe.hh"

#define MAXZEILENLAENGE  2000

#define DEFSUBSTMODEL JC69
#define DEFINDMODEL TKF1
#define DEFINFILE "infile"
#define DEFFORMAT 0
#define DEFGUIDETREE "guidetree"
#define DEFOUTTREE "outtree"
#define DEFOUTALIGN "outalign"
#define DEFOUTPARAM "outparam"
#define DEFLOGFILE "logfile"
#define DEFDELTA 1.0
#define DEFCOOLALI 1.00001
#define DEFCTREE 0.5
#define DEFCPARAM 0.5
#define DEFALPHATREE 0.5
#define DEFALPHAPARAM 0.5
#define DEFMAXSTEPS  ULONG_MAX
#define DEFRESAMPRANGE 50
#define DEFREPORTAFTER 1000
#define DEFBURNIN 1000

eingabe_parameter::eingabe_parameter()
{
}

void eingabe_parameter::from_file(char * dateiname)
    /*
     * Eingabe der eingabe_parameter aus Datei dateiname 
     */
{
    scriptpart alt=scNONE;
    scriptpart aktuell=scNONE;
    subst_option so;
    indel_option io;
    files_option fo;
    keep_option ko;
    cool_option co;
    option_arg oa;
    string argument, Zeile;
    char *cZeile;
    keep_intermediate kint;
    
    ifstream inputstr(dateiname);
    if (!inputstr) {
	cerr << "cannot open file " << dateiname << endl;
	exit(EXIT_FAILURE);
    }
    if((cZeile=(char *)malloc(MAXZEILENLAENGE))==NULL) {
       cerr << "Cannot allocate buffer for the control file.\n";
       exit(EXIT_FAILURE);
    }
    while(!inputstr.eof()) {
        inputstr.getline(cZeile,MAXZEILENLAENGE);
	Zeile = string(cZeile);
        aktuell = isnewpart(Zeile);
        if(alt!=aktuell && aktuell!=scNONE)
	   alt=aktuell;
	else {
           switch(alt) {
               case scNONE :
	          break;
	       case SUBST :
	          so = whichsubstoption(Zeile,argument,oa);
	          if(so!=suNONE)
		     sub.adjust(so,argument,oa);
	          break;
	       case INDEL :
	          io = whichindeloption(Zeile,argument,oa);
	          if(io!=inNONE)
		     ind.adjust(io,argument,oa);
	          break;
               case FILES :
	          fo = whichfilesoption(Zeile,argument,oa);
	          if(fo!=fiNONE)
	             fil.adjust(fo,argument);
                  break;
	       case KEEP :
	          ko = whichkeepoption(Zeile,argument,oa);
	          if(ko!=keNONE)
	             kint.adjust(ko,argument,oa);
                  break;
	       case COOL :
	          co = whichcooloption(Zeile,argument,oa);
	          if(co!=coNONE)
	             coo.adjust(co,argument,oa);
                  break;
               default :
	          cerr << "Unknown script part\n";
		  exit(EXIT_FAILURE);
           }
       }
    }
    inputstr.close();
    
    sub.abgleich();  // passt auch alles?
    ind.abgleich();
    kee=keepfixed(sub.model,ind.model,kint);
    
}



// substmod_parameters enthaelt die Einstellungen,
// die man fuer das Substitutionsmodell gewaehlt hat.
// model = Typ des Substitutionsmodells
// parameter = Parameter des Modells
// freqs = Zustandshaeufigkeiten
// freqset = sollen Basenhaeufigkeiten geschaetzt werden 
// Die Klasse substmod_parameters wird benutzt, um das in
// der Klasse mult_alignment enthaltene Substitutionsmodell
// zu konstruieren. 

substmod_parameters::substmod_parameters() // default constructor
{
   model = DEFSUBSTMODEL;
   modelset = FALSE;
   paramset = FALSE;
   freqset = FALSE;
}

void substmod_parameters::adjust(subst_option so,string argument,option_arg oa)
{
   switch(so) {
       case SMODEL :
            setmodel(argument,oa);
	    break;
       case SPARAM :
            setparams(argument,oa);
	    break;
       case FREQS :
            setfreqs(argument,oa);
	    break;
       default :
           cerr << "Impossible option "<<(int)so<<" under headline SUBST\n";
	   exit(EXIT_FAILURE);
   }
}

void substmod_parameters::setmodel(string argument,option_arg oa)
{

   if(modelset) {
      cerr << "\nWARNING:\n"
           << "The substitution model has already been set.\n"
           << "The older value is now overwritten.\n\n";
   }
   modelset = TRUE;
   switch(oa) {
       case VALUE :
            model = stringtosubarch(argument);
            break;
       case DEFAULT :
	    model = DEFSUBSTMODEL;
            break;
       default :
            cerr << "Invalid argument " << oatostring(oa) << " to option model\n";
	    exit(EXIT_FAILURE);
   }
}

void substmod_parameters::setparams(string argument,option_arg oa)
{
   if(paramset) {
      cerr << "\nWARNING:\n"
           << "The substitution model's parameters have already been set.\n"
           << "The older values are now overwritten.\n\n";
   }
   switch(oa){
       case DEFAULT :
            paramset = FALSE;
            break;
       case VECTOR :
            paramset = TRUE;
            stringtovector(argument,params);
            break;
       default :
            cerr << "Invalid argument " << oatostring(oa) << " to option parameter\n";
	    exit(EXIT_FAILURE);
   }
}

void substmod_parameters::setfreqs(string argument,option_arg oa)
{
   if(freqset) {
      cerr << "\nWARNING:\n"
           << "The base or amino acid frequencies have already been set.\n"
           << "The older values are now overwritten.\n\n";
   }
   switch(oa){
       case ESTIMATE :
       case DEFAULT :
            freqset = FALSE;
            break;
       case VECTOR :
            freqset = TRUE;
            stringtovector(argument,freqs);
            break;
       default :
            cerr << "Invalid argument " << oatostring(oa) << " to option freqs\n";
	    exit(EXIT_FAILURE);
   }
}

void substmod_parameters::abgleich()
{
   unsigned int i;
   bool flag;
   bool viertelflag;
   
   if(modelset) {
     
     // Welche Wahl wurde fuer die Subst.raten getroffen?
     switch(model) {
          case FEL81 :
          case JC69 :
             if(paramset && params.size()>0) {
                cerr << "The Felsenstein model and the Jukes Cantor do not use\n"
	             << "any rate parameters.\n";
		exit(EXIT_FAILURE);
	     }
	     params.clear();
             break;
          case HKY :
          case KIMURA2P :
	     if(paramset) {  // Userwunsch
	          if(params.size()!=1) {
                    cerr << "The HKY and the K2P model have only one rate parameter.\n";
                    exit(EXIT_FAILURE);
                  } else {
		    if(params[0]<=0.0) {
		         cerr << "The parameters of the substitution model\n"
			      << "have to be positive.\n";
			 exit(EXIT_FAILURE);
		    }
		  }
	     } else {  // kein Userwunsch
		  params.clear(); 
                  params.push_back(0.5);
	     }
	     break;
          case GENREV :
	     if(paramset) {  // Userwunsch
	          if(params.size()!=6) {
                    cerr << "The general reversible model has six rate parameters.\n";
                    exit(EXIT_FAILURE);
                  } else {
		    flag = TRUE;
		    for(i=0;i<params.size();i++)
		      flag = flag && (params[i]>0.0);
		    if(!flag) {
		         cerr << "The parameters of the substitution model\n"
			      << "have to be positive.\n";
			 exit(EXIT_FAILURE);
		    }
		  }
	     } else {  // kein Userwunsch
		  params.clear();
		  for(i=0;i<6;i++) 
                    params.push_back(1.0);
	     }
	     break;
          case TAMNEI :
          case KIMURA3P :
	     if(paramset) {  // Userwunsch
	          if(params.size()!=2) {
                    cerr << "The Tamura-Nei and the Kimura3P model have two rate parameters.\n";
                    exit(EXIT_FAILURE);
                  } else {
		    flag = TRUE;
		    for(i=0;i<params.size();i++)
		      flag = flag && (params[i]>0.0);
		    if(!flag) {
		         cerr << "The parameters of the substitution model\n"
			      << "have to be positive.\n";
			 exit(EXIT_FAILURE);
		    }
		  }
	     } else {  // kein Userwunsch
		  params.clear();
                  params.push_back(0.5);
                  params.push_back(1.0);
	     }
             break;
          case MV2000 :
	     if(paramset) {  // Userwunsch
	          if(params.size()!=190) {
                    cerr << "The general reversible model for proteins has 190 rate parameters.\n";
                    exit(EXIT_FAILURE);
                  } else {
		    flag = TRUE;
		    for(i=0;i<params.size();i++)
		      flag = flag && (params[i]>0.0);
		    if(!flag) {
		         cerr << "The parameters of the substitution model\n"
			      << "have to be positive.\n";
			 exit(EXIT_FAILURE);
		    }
		  }
	     } else {  // kein Userwunsch
		  params.clear();
                  for(i=0;i<190;i++)
                    params.push_back(1.0);
	     }
             break;
          case DAYHOFF :
          case BLOSUM62 :
          case JTT :
          case MTREV24 :
	     if(paramset || freqset) {
	        cerr << "The substitution model that you want to use does not\n"
		     << "accept user specified rate parameters or amino acid\n"
		     << "frequencies.\n";
		exit(EXIT_FAILURE);
	     }
             break;   
          default :
	     cerr << "Unknown substitution model\n";
	     exit(EXIT_FAILURE);
     }
     
     // Welche Wahl wurde fuer die Zustandshaeufigkeiten getroffen?
     if(freqset) { // Wenn nichts angegeben wurde, wird eh geschaetzt
         flag=TRUE;
	 for(i=0;i<freqs.size();i++)
	     flag = flag && (freqs[i]>0.0);
	 if(!flag) {
	    cerr << "All nucleotide or amino acid frequencies have to be positive.\n";
	    exit(EXIT_FAILURE);
	 }
	 // Ob die Summe der Haeufigkeiten 1 ist, wird nicht ueberprueft,
	 // da der Konstruktor von substmod mit absoluten Haeufigkeiten
	 // umgehen kann
         switch(model) {
              case JC69 :
              case KIMURA2P :
              case KIMURA3P :
              case DAYHOFF :
              case BLOSUM62 :
              case JTT :
              case MTREV24 :
	         cerr << "The substitution model that you want to use does not\n"
		      << "accept user specified rate parameters or state frequencies.\n";
		 exit(EXIT_FAILURE);
                 break; 
              case FEL81 :
	         if(freqs.size()!=4) {
		    cerr << "Please specify exactly four frequencies for A,C,G and T.\n";
		    exit(EXIT_FAILURE);
		 }
		 viertelflag=TRUE;
		 for(i=0;i<freqs.size();i++)
		     viertelflag=viertelflag && (freqs[i]>0.25-DBL_MIN && freqs[i]<0.25+DBL_MIN);
		 if(viertelflag) {
		    cerr << "You want all base frequencies to be equal,\n"
		         << "so the FEL81 model becomes the JC69 model.\n";
		    model = JC69;
		 }
                 break;
              case HKY :
	         if(freqs.size()!=4) {
		    cerr << "Please specify exactly four frequencies for A,C,G and T.\n";
		    exit(EXIT_FAILURE);
		 }
		 viertelflag=TRUE;
		 for(i=0;i<freqs.size();i++)
		     viertelflag=viertelflag && (freqs[i]>0.25-DBL_MIN && freqs[i]<0.25+DBL_MIN);
		 if(viertelflag) {
		    cerr << "You want all base frequencies to be equal,\n"
		         << "so the HKY model becomes the KIMURA2P model.\n";
		    model = KIMURA2P;
		 }
	         break;
              case GENREV :
	         if(freqs.size()!=4) {
		    cerr << "Please specify exactly four frequencies for A,C,G and T.\n";
		    exit(EXIT_FAILURE);
		 }
		 break;
              case TAMNEI :
	         if(freqs.size()!=4) {
		    cerr << "Please specify exactly four frequencies for A,C,G and T.\n";
		    exit(EXIT_FAILURE);
		 }
		 viertelflag=TRUE;
		 for(i=0;i<freqs.size();i++)
		     viertelflag=viertelflag && (freqs[i]>0.25-DBL_MIN && freqs[i]<0.25+DBL_MIN);
		 if(viertelflag) {
		    cerr << "You want all base frequencies to be equal,\n"
		         << "so the Tamura-Nei model becomes the KIMURA3P model.\n";
		    model = KIMURA3P;
		 }
	         break;
              case MV2000 :
	         if(freqs.size()!=20) {
		    cerr << "Please specify exactly twenty amino acid frequencies.\n";
		    exit(EXIT_FAILURE);
		 }
		 break;
              default :
	         cerr << "Unknown substitution model\n";
	         exit(EXIT_FAILURE);
         }
     }

   } else {
       if(paramset) {
          cerr << "Please specify a substitution model using the option 'model=X'\n"
	       << "Possible choices for X are:\n"
	       << "for nucleotides:             for amino acids:\n"
	       << "   FEL81                        MV2000\n"
	       << "   JC69                         DAYHOFF\n"
	       << "   HKY                          BLOSUM62\n"
	       << "   KIMURA2P                     JTT\n"
	       << "   TAMNEI                       MTREV24\n"
	       << "   KIMURA3P\n"
	       << "   GENREV\n\n";
	  exit(EXIT_FAILURE);
       } else {
           if(freqset) {
              cerr << "Please specify a substitution model using the option 'model=X'\n"
	           << "Possible choices for X are:\n"
	           << "for nucleotides:             for amino acids:\n"
	           << "   FEL81                        MV2000\n"
	           << "   JC69                         DAYHOFF\n"
	           << "   HKY                          BLOSUM62\n"
	           << "   KIMURA2P                     JTT\n"
	           << "   TAMNEI                       MTREV24\n"
	           << "   KIMURA3P\n"
	           << "   GENREV\n\n";
	       exit(EXIT_FAILURE);	   
	   } 
	   else
	       model=JC69;
       }
   }
}



// indel_parameters enthaelt die Einstellungen fuer das Insertions-Deletionsmodell.
// Analog zu substmod_parameters
indel_parameters::indel_parameters()
{
   model = DEFINDMODEL;
   modelset = FALSE;
   paramset = FALSE;
   switch(model) {
     case TKF1 :
       params.push_back(0.01);
       params.push_back(0.011);
       break;
     case TKF2 :
       params.push_back(0.01);
       params.push_back(0.011);
       params.push_back(0.5);   // Parameter der Fragmentlaenge
       break;
     default :
       cerr << "At the moment there is only TKF1 and TKF2\n";
       exit(EXIT_FAILURE);
   }
}

void indel_parameters::adjust(indel_option io,string argument,option_arg oa)
{
   switch(io) {
       case IMODEL :
            setmodel(argument,oa);
	    break;
       case IPARAM :
            setparams(argument,oa);
	    break;
       default :
           cerr << "Impossible option under headline INDEL\n";
	   exit(EXIT_FAILURE);
   }
}

void indel_parameters::setmodel(string argument,option_arg oa)
{

   if(modelset) {
      cerr << "\nWARNING:\n"
           << "The insertion deletion model has already been set.\n"
           << "The older value is now overwritten.\n\n";
   }
   modelset = TRUE;
   switch(oa) {
       case VALUE :
            model = stringtoidarch(argument);
            break;
       case DEFAULT :
	    model = DEFINDMODEL;
            break;
       default :
            cerr << "Invalid argument " << oatostring(oa) << " to option model\n";
	    exit(EXIT_FAILURE);
   }
}

void indel_parameters::setparams(string argument,option_arg oa)
{
   if(paramset) {
      cerr << "\nWARNING:\n"
           << "The insertion deletion model's parameters have already been set.\n"
           << "The older values are now overwritten.\n\n";
   }
   switch(oa){
       case DEFAULT :
            paramset = FALSE;
            break;
       case VECTOR :
            paramset = TRUE;
            stringtovector(argument,params);
            break;
       default :
            cerr << "Invalid argument " << oatostring(oa) << " to option parameter\n";
	    exit(EXIT_FAILURE);
   }
}

void indel_parameters::abgleich()
{
    if(modelset) {
      switch(model) {
         case TKF1 :
	    if(paramset) {
	      if(params.size()!=2) {
	         cerr << "The TKF1 model has two parameters.\n";
		 exit(EXIT_FAILURE);
	      }
	      if(params[0]<=0.0 || params[1]<=0.0 || params[1]<=params[0]) {
	         cerr << "Both rates of the TKF1 model have to be positive and\n"
		      << "the first rate has to be smaller than the second one.\n";
		 exit(EXIT_FAILURE);
	      }
	    } else {
	      params.clear();
              params.push_back(0.01);
              params.push_back(0.011);
	    }
	    break;
         case TKF2 :
	    if(paramset) {
	      if(params.size()!=3) {
	         cerr << "The TKF2 model has three parameters.\n";
		 exit(EXIT_FAILURE);
	      }
	      if(params[0]<=0.0 || params[1]<=0.0 || params[1]<=params[0]) {
	         cerr << "Both rates of the TKF2 model have to be positive and\n"
		      << "the first rate has to be smaller than the second one.\n";
		 exit(EXIT_FAILURE);
	      }
	      if(params[2]<=0.0 || params[2]>=1.0) {
	         cerr << "The third parameter of the TKF2 model should be positive\n"
		      << "and smaller than 1.\n";
		 exit(EXIT_FAILURE);
	      }
	    } else {
	      params.clear();
              params.push_back(0.01);
              params.push_back(0.011);
              params.push_back(0.5);
	    }
	    break;
         default :
	    cerr << "Unknown insertion and deletion model\n";
	    exit(EXIT_FAILURE);
      }
    } else {
      if(paramset) {
        cerr << "Please specify an insertion and deletion model.\n"
	     << "Possible choices are:\n"
	     << "     TKF1\n"
	     << "     TKF2\n\n";
	exit(EXIT_FAILURE);
      } else {
        model = TKF1;
	params.clear();
        params.push_back(0.01);
        params.push_back(0.011);
      }
    }
}

// filenames enthaelt die gewuenschten Dateinamen
filenames::filenames()
{
   infile=DEFINFILE;
   format=DEFFORMAT; 
   guidetree=DEFGUIDETREE; 
   outtree=DEFOUTTREE; 
   outalign=DEFOUTALIGN;
   outparam=DEFOUTPARAM;
   logfile=DEFLOGFILE;
}

void filenames::adjust(files_option fo, string argument)
{
   switch(fo) {
      case INFILE :
         infile=argument; //whichfilesoption sollte schon den richtigen Teil extrahiert haben
         break;
      case FORMAT :
         if(sscanf(argument.c_str(),"%d",&format)!=1) {
	    cerr << "Cannot scan from the argument of 'format = '\n";
	    exit(EXIT_FAILURE);
	 }
         break;
      case GUIDETREE :
         guidetree=argument;
         break;
      case OUTTREE :
         outtree=argument;
         break;
      case OUTALIGN :
         outalign=argument;
         break;
      case OUTPARAM :
         outparam=argument;
	 break;
      case LOGFILE :
         logfile=argument;
         break;
      default :
         cerr << "Impossible option under headline FILES\n";
	 exit(EXIT_FAILURE);
   }
}

// keepfixed erfasst, welche Modellparameter nicht geaendert
// werden sollen
keepfixed::keepfixed()
{
}

keepfixed::keepfixed(subarch sub, idarch id)
{
   unsigned int i;
   switch(sub) {
      case GENREV :
         subfix.resize(6);
	 break;
      case HKY :
      case KIMURA2P :
         subfix.resize(1);
	 break;
      case TAMNEI :
      case KIMURA3P :
         subfix.resize(2);
	 break;
      case MV2000 :
         subfix.resize(190);
	 break;
      case FEL81 :
      case JC69 :
      case JTT :
      case DAYHOFF :
      case MTREV24 :
      case BLOSUM62 :
         break;
      default :
        cerr << "Unknown substitution model in function keepfixed.\n";
	exit(EXIT_FAILURE);
   }
   switch(id) {
      case TKF1 :
        indfix.resize(2);
        break;
      case TKF2 :
        indfix.resize(3);
        break;
      default :
        cerr << "Unknown indel model in function keepfixed.\n";
	exit(EXIT_FAILURE);
   }
   for(i=0;i<subfix.size();i++)
       subfix[i]=FALSE;
   for(i=0;i<indfix.size();i++)
       indfix[i]=FALSE;
}

keepfixed::keepfixed(subarch sub, idarch id, keep_intermediate kint)
{
   unsigned int i;
   switch(sub) {
      case GENREV :
         subfix.resize(6);
	 break;
      case HKY :
      case KIMURA2P :
         subfix.resize(1);
	 break;
      case TAMNEI :
      case KIMURA3P :
         subfix.resize(2);
	 break;
      case MV2000 :
         subfix.resize(190);
	 break;
      case FEL81 :
      case JC69 :
      case JTT :
      case DAYHOFF :
      case MTREV24 :
      case BLOSUM62 :
         break;
      default :
        cerr << "Unknown substitution model in function keepfixed.\n";
	exit(EXIT_FAILURE);
   }
   switch(id) {
      case TKF1 :
        indfix.resize(2);
        break;
      case TKF2 :
        indfix.resize(3);
        break;
      default :
        cerr << "Unknown indel model in function keepfixed.\n";
	exit(EXIT_FAILURE);
   }
   if(kint.subfix.size()>subfix.size() || kint.indfix.size()>indfix.size()) {
      cerr << "You want to keep more parameters fixed than possible\n";
      exit(EXIT_FAILURE);
   }
   if(kint.subfixset && kint.subfix.size()==0) { // entspricht substmod[ALL]
      for(i=0;i<subfix.size();i++)
          subfix[i]=TRUE;
   } else {
      for(i=0;i<kint.subfix.size();i++)
          subfix[i]=kint.subfix[i];
      while(i<subfix.size()) {
          subfix[i]=FALSE;
	  i++;
      }
   }
   if(kint.indfixset && kint.indfix.size()==0) { // entspricht indmod[ALL]
      for(i=0;i<indfix.size();i++)
          indfix[i]=TRUE;
   } else {
      for(i=0;i<kint.indfix.size();i++)
          indfix[i]=kint.indfix[i];
      while(i<indfix.size()) {
          indfix[i]=FALSE;
	  i++;
      }
   }
   
}


keepfixed keepfixed::operator=(keepfixed kee2)
{
   subfix=kee2.subfix;
   indfix=kee2.indfix;
   
   return *this;
}


// keep_intermediate vermittelt zwischen dem Steuerungsscript und keepfixed

keep_intermediate::keep_intermediate()
{
   subfixset=FALSE;
   indfixset=FALSE;
}

void keep_intermediate::adjust(keep_option ko, string argument, option_arg oa)
{
   list<int> buffer;
   unsigned int i, lsize;
   int alt, j;
   
   switch(ko) {
       case SUBSTMOD :
          switch(oa) {
	     case ALL :
	         subfix.clear(); 
		 subfixset=TRUE; 
	         break;
	     case INTERVAL :
	         stringtolist(argument,buffer);
		 buffer.sort();
		 alt=-1;
		 subfix.clear();
		 lsize=buffer.size();
		 for(i=0;i<lsize;i++) {
		    for(j=alt+1;j<buffer.front();j++)  // direkter Elementzugriff ist bei Listen nicht mgl.
		        subfix.push_back(FALSE);
		    subfix.push_back(TRUE);
		    buffer.pop_front();
		    alt=(int)i; 
		 }
		 subfixset=TRUE;
	         break;
	     default :
	         cerr << "Invalid argument to option substmod\n";
		 exit(EXIT_FAILURE);
	  }
          break;
       case INDMOD :
          switch(oa) {
	     case ALL :
	         indfix.clear();
		 indfixset=TRUE;
	         break;
	     case INTERVAL :
	         stringtolist(argument,buffer);
		 buffer.sort();
		 alt=-1;
		 indfix.clear();
		 lsize=buffer.size();
		 for(i=0;i<lsize;i++) {
		    for(j=alt+1;j<buffer.front();j++)
		        indfix.push_back(FALSE);
		    indfix.push_back(TRUE);
		    buffer.pop_front();
		    alt=(int)i; 
		 }
		 indfixset=TRUE;
	         break;
	     default :
	         cerr << "Invalid argument to option indmod\n";
		 exit(EXIT_FAILURE);
	  }
          break;
       default :
           cerr << "Impossible option under headline KEEP\n";
	   exit(EXIT_FAILURE);
   }
}

// cooling enthaelt Werte, die fuer das Sampeln wichtig sind
cooling::cooling()
{
  delta=DEFDELTA;      // Welche Werte hier sinnvoll sind, weiss ich nicht.
  coolali=DEFCOOLALI;
  ctree=DEFCTREE;
  cparam=DEFCPARAM;
  alphatree=DEFALPHATREE;
  alphaparam=DEFALPHAPARAM;
  burn_in=DEFBURNIN;
  maxsteps=DEFMAXSTEPS;
  reportafter=DEFREPORTAFTER;
  resamprange=DEFRESAMPRANGE;
}


void cooling::adjust(cool_option co, string argument, option_arg oa)
{
   switch(co) {
       case DELTA :
          setdelta(argument,oa);
          break;
       case COOLALI :
          setcoolali(argument,oa);
          break;
       case CTREE :
          setctree(argument,oa);
          break;
       case CPARAM :
          setcparam(argument,oa);
          break;
       case ALPHATREE :
          setalphatree(argument,oa);
          break;
       case ALPHAPARAM :
          setalphaparam(argument,oa);
          break;
       case BURNIN :
          setburnin(argument,oa);
          break;
       case MAXSTEPS :
          setmaxsteps(argument,oa);
          break;
       case REPORTAFTER :
          setreportafter(argument,oa);
          break;
       case RESAMPRANGE :
          setresamprange(argument,oa);
          break;
       default :
           cerr << "Impossible option under headline COOL\n";
	   exit(EXIT_FAILURE);
   }
}

void cooling::setdelta(string argument, option_arg oa)
{
   vector<double> buffer;
   switch(oa) {
       case VALUE :
          stringtovector(argument,buffer);
	  if(buffer.size()!=1) {
	     cerr<<"Internal error in function setdelta\n";
	     exit(EXIT_FAILURE);
	  }
	  delta=buffer[0];
          break;
       case DEFAULT :
          delta=DEFDELTA;
	  break;
       default :
            cerr << "Invalid argument " << oatostring(oa) << " to option delta\n";
	    exit(EXIT_FAILURE);
   }
   if(delta<=0.0) {
      cerr << "Delta has to be positive!\n";
      exit(EXIT_FAILURE);
   }
}

void cooling::setcoolali(string argument, option_arg oa)
{
   vector<double> buffer;
   switch(oa) {
       case VALUE :
          stringtovector(argument,buffer);
	  if(buffer.size()!=1) {
	     cerr<<"Internal error in function setdelta\n";
	     exit(EXIT_FAILURE);
	  }
	  coolali=buffer[0];
          break;
       case DEFAULT :
          coolali=DEFCOOLALI;
	  break;
       default :
            cerr << "Invalid argument " << oatostring(oa) << " to option coolali\n";
	    exit(EXIT_FAILURE);
   }
   if(coolali<=1.0) {
      cerr << "Coolali has to be bigger than 1.0\n";
      exit(EXIT_FAILURE);
   }
   if(coolali>1.001)
      cerr << "WARNING: coolali should only be slightly bigger than 1.0\n";
}

void cooling::setctree(string argument, option_arg oa)
{
   vector<double> buffer;
   switch(oa) {
       case VALUE :
          stringtovector(argument,buffer);
	  if(buffer.size()!=1) {
	     cerr<<"Internal error in function setdelta\n";
	     exit(EXIT_FAILURE);
	  }
	  ctree=buffer[0];       
          break;
       case DEFAULT :
          ctree=DEFCTREE;
	  break;
       default :
            cerr << "Invalid argument " << oatostring(oa) << " to option ctree\n";
	    exit(EXIT_FAILURE);
   }
   if(ctree<=0.0) {
      cerr << "ctree has to be bigger than 0.0\n";
      exit(EXIT_FAILURE);
   }
}

void cooling::setcparam(string argument, option_arg oa)
{
   vector<double> buffer;
   switch(oa) {
       case VALUE :
          stringtovector(argument,buffer);
	  if(buffer.size()!=1) {
	     cerr<<"Internal error in function setdelta\n";
	     exit(EXIT_FAILURE);
	  }
	  cparam=buffer[0];       
          break;
       case DEFAULT :
          cparam=DEFCPARAM;
	  break;
       default :
            cerr << "Invalid argument " << oatostring(oa) << " to option cparam\n";
	    exit(EXIT_FAILURE);
   }
   if(cparam<=0.0) {
      cerr << "cparam has to be bigger than 1.0\n";
      exit(EXIT_FAILURE);
   }
}

void cooling::setalphatree(string argument, option_arg oa)
{
   vector<double> buffer;
   switch(oa) {
       case VALUE :
          stringtovector(argument,buffer);
	  if(buffer.size()!=1) {
	     cerr<<"Internal error in function setdelta\n";
	     exit(EXIT_FAILURE);
	  }
	  alphatree=buffer[0];       
          break;
       case DEFAULT :
          alphatree=DEFALPHATREE;
	  break;
       default :
            cerr << "Invalid argument " << oatostring(oa) << " to option alphatree\n";
	    exit(EXIT_FAILURE);
   }
   if(alphatree<=0.0) {
      cerr << "alphatree has to be bigger than 0.0\n";
      exit(EXIT_FAILURE);
   }
}

void cooling::setalphaparam(string argument, option_arg oa)
{
   vector<double> buffer;
   switch(oa) {
       case VALUE :
          stringtovector(argument,buffer);
	  if(buffer.size()!=1) {
	     cerr<<"Internal error in function setdelta\n";
	     exit(EXIT_FAILURE);
	  }
	  alphaparam=buffer[0];       
          break;
       case DEFAULT :
          alphaparam=DEFALPHAPARAM;
	  break;
       default :
            cerr << "Invalid argument " << oatostring(oa) << " to option alphaparam\n";
	    exit(EXIT_FAILURE);
   }
   if(alphaparam<=0.0) {
      cerr << "alphaparam has to be bigger than 1.0\n";
      exit(EXIT_FAILURE);
   }
}

void cooling::setburnin(string argument, option_arg oa)
{
   vector<unsigned long int> buffer;
   switch(oa) {
       case VALUE :
          stringtovector(argument,buffer);
	  if(buffer.size()!=1) {
	     cerr<<"Internal error in function setdelta\n";
	     exit(EXIT_FAILURE);
	  }
	  burn_in=buffer[0];       
          break;
       case DEFAULT :
          burn_in=DEFBURNIN;
	  break;
       default :
            cerr << "Invalid argument " << oatostring(oa) << " to option burn_in\n";
	    exit(EXIT_FAILURE);
   }
   if(burn_in<=1) {
      cerr << "You should allow more than one burn in steps!\n";
      exit(EXIT_FAILURE);
   }
}

void cooling::setmaxsteps(string argument, option_arg oa)
{
   vector<unsigned long int> buffer;
   switch(oa) {
       case VALUE :
          stringtovector(argument,buffer);
	  if(buffer.size()!=1) {
	     cerr<<"Internal error in function setdelta\n";
	     exit(EXIT_FAILURE);
	  }
	  maxsteps=buffer[0];       
          break;
       case DEFAULT :
          maxsteps=DEFMAXSTEPS;
	  break;
       default :
            cerr << "Invalid argument " << oatostring(oa) << " to option maxsteps\n";
	    exit(EXIT_FAILURE);
   }
}

void cooling::setreportafter(string argument, option_arg oa)
{
   vector<unsigned long int> buffer;
   switch(oa) {
       case VALUE :
          stringtovector(argument,buffer);
	  if(buffer.size()!=1) {
	     cerr<<"Internal error in function setdelta\n";
	     exit(EXIT_FAILURE);
	  }
	  reportafter=buffer[0];       
          break;
       case DEFAULT :
          reportafter=DEFREPORTAFTER;
	  break;
       default :
            cerr << "Invalid argument " << oatostring(oa) << " to option reportafter\n";
	    exit(EXIT_FAILURE);
   }
}

void cooling::setresamprange(string argument, option_arg oa)
{
   vector<unsigned long int> buffer;
   switch(oa) {
       case VALUE :
          stringtovector(argument,buffer);
	  if(buffer.size()!=1) {
	     cerr<<"Internal error in function setdelta\n";
	     exit(EXIT_FAILURE);
	  }
	  resamprange=buffer[0];       
          break;
       case DEFAULT :
          resamprange=DEFRESAMPRANGE;
	  break;
       default :
            cerr << "Invalid argument " << oatostring(oa) << " to option resamprange\n";
	    exit(EXIT_FAILURE);
   }
}

cooling cooling::operator=(cooling coo2)
{
  delta=coo2.delta;
  coolali=coo2.coolali;
  ctree=coo2.ctree;
  cparam=coo2.cparam;
  alphatree=coo2.alphatree;
  alphaparam=coo2.alphaparam;
  burn_in=coo2.burn_in;
  maxsteps=coo2.maxsteps;
  reportafter=coo2.reportafter;
  resamprange=coo2.resamprange;
  
  return *this;
}
