/****************************************************************************
 * This file is part of the program ALIFRITZ                                *
 *                                                                          *
 *    Copyright (C) 2001  Roland Fleissner                                  *
 *    e-mail: fleissne@uidaho.edu                             		    * 
 *                                                                          *
 *  This program is free software; you can redistribute it and/or modify    *
 *  it under the terms of the GNU General Public License as published by    *
 *  the Free Software Foundation; either version 2 of the License, or       *
 *  (at your option) any later version.                                     *
 *                                                                          *
 *  This program is distributed in the hope that it will be useful,         *
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of          *
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           *
 *  GNU General Public License for more details.                            *
 *                                                                          *
 *  You should have received a copy of the GNU General Public License       *
 *  along with this program; if not, write to the                           *
 *  Free Software Foundation, Inc., 59 Temple Place, Suite 330,             *
 *  Boston, MA  02111-1307  USA                                             *
 ****************************************************************************/

#ifndef MULT_H
#define MULT_H


enum swapflag {WggY=1,WggZ=2,XggY=3,XggZ=4};
// gibt an, welche Aeste des Quartetts WX||YZ ausgetauscht werden

class mercedes_stern { // nur fuer die geordnete Uebergabe einiger Parameter
public:
   baum *tree;
   vector<quart_zus> ali;
   vector<unsigned int> ordered_children;
   int verstossen;
   mercedes_stern();
   mercedes_stern(baum *p, vector<quart_zus>& vieral, vector<unsigned int>& kd, int childtochange);
};


class mdataset : public dataset {
public:
   mdataset();
   mdataset(baum *b, alphsize alf);
   void operator=(mdataset md2);
   void row(unsigned int i, vector<char>& rowvec);  // schreibt Spalte i nach rowvec
   void push_mult(baum *b, vector<int>& ali_to_node, vector<int>& node_to_ali,
                  vector<vector<char> >& idmat, vector<vector<char> >& seqmat);
   bool validcolumn(vector<char>& col, unsigned int expsize);
   bool notatend(vector<zustand>& pzus, bool& syncstart);
};

class compact_subst_pattern {
public:
   vector<vector<char> > muster; // Positionsmuster (sortiert)
   vector<int> anzmus;  // Wie viele gibt's von jedem Muster
   vector<string> ids;  // Taxonnamen
   
   compact_subst_pattern();
   compact_subst_pattern(mdataset &pali);
   ~compact_subst_pattern();
   void operator=(compact_subst_pattern csp2);
};


class mult_alignment {
private :
   /**** Variablen   ******/
   substmodel sub;
   indel id;
   baum *told;
   baum *tnew;
   keepfixed kee;
   cooling coo;
   unsigned long int zaehler;
   mdataset printform;
   compact_subst_pattern cosubi; // Welche Substitutionsmuster findet man in printform?
   double lnL; // log-Likelihood
   double maxdifftree; // maximale Differenz der logLikelihhoods beim Baeume sampeln im Burn-in
   double maxdiffparam; // maximale Differenz der Bewertungsfunktion beim Parameter sampeln im Burn-in
   double betatree;    // Simulated-Annealing-Parameter beta fuers Baeumesampeln
   double betaparam;    // Simulated-Annealing-Parameter beta fuers Parametersampeln
   unsigned int deltafaktor;  // Zaehler des Faktors mit dem MN-delta multipliziert wird
   unsigned int MNstep;   // Stufe der delta-Kalibrierung
   bool descending;
   vector<double> wieweit;
   vector<int> times_visited;

   /**** Funktionen   ******/
   
   /*********   Konstruktion und Optimierung   ******************/
   void compute_distmat(dataset& d, vector<vector<double> >& distmat);
   double median_scale(vector<vector<double> >& scalemat);
   // Median der Eintraege in der rechten oberen Haelfte von scalemat
   void construct_align(baum *tree);
          // konstruiere erstes Alignment
   void midseq(baum* p, vector<unsigned int>& Kinder);
          // zieht vorwaertsgehend ein Dreieralignment
   bool uphill(baum *tree);            
          // optimieren des Alignments
   bool uphill(baum *tree, substmodel& submod, indel& indmod);            
          // optimieren des Alignments
   bool trytoimprove();
          // probiert ob man mit uphill zu einem besseren Alignment kommt
   void samp_centseq(baum *b, vector<unsigned int>& Kinder, substmodel& submod, indel& indmod);
          // erzeugt ein b->pro, gegeben das Alignment der Nachbarn
   void central_pair(mercedes_stern& m1, mercedes_stern& m2);
          // erzeugt alignierte Sequenzen an den Knoten m1.tree und m2.tree
   bool alteroderneuerbaum();
          // waehlt zwischen altem und neuem Baum


   /*********   berechnen der Likelihood   ******************/
   double complik(compact_subst_pattern& csb, baum *tree); 
          // logLikelihood des Alignments
   double complik(compact_subst_pattern& csb, baum *tree, indel& newid, substmodel& newsub); 
          // logLikelihood des Alignments fuer die angegebenen Indel- und Substitutionsmodelle
   double indlik(baum *tree, indel& newid); 
          // Insertions-Deletions-Summand der logLikelihood des Alignments fuer das Indelmodell newid
   double sublik(compact_subst_pattern& csb, baum *tree, substmodel& newsub); 
          // Substitutions-Summand der logLikelihood des Alignments fuer das Subst.modell newsub
   double logprobofpattern(vector<char>& pattern, baum *tree, substmodel& newsub);
   void   childliks(vector<char>& pattern, baum *tree, substmodel& newsub, vector<double>& logliks, int alf);


   /**********  Verschiedenes      ************************/
   void familie(vector<unsigned int>& stern, baum *p);
          // Was hat p fuer Nachbarn
   void familie(vector<unsigned int>& stern, baum *p, baum *q);
          // ist q nicht der Vater wird's das rechte Kind
   void ausaltmachneu();  // kopiert told nach tnew
   void ausneumachalt();  // kopiert tnew nach told
   void keeptrack(vector<baum *>& nodes, vector<unsigned int>& nbidx, baum *p, int ownch, baum *q, int nbch);

   /**********  zurueckschreiben in den Baum  ************************/
   void write_midseq_to_tree(vector<triplet_zus>& dreiseq, baum *p, vector<unsigned int>& Kinder);
          // weist dem mittleren Knoten ein Profil zu
	  // und setzt alle paarweisen Alignments auf die richtigen Werte
   void write_centseq(vector<quart_zus>& quart, baum *b, vector<unsigned int>& Kinder, substmodel& submod);
          // Schreibt die gezogene Sequenz zurueck in den Baum
   void write_centpair(vector<hex_zus>& allesequs, mercedes_stern& anc, mercedes_stern& rec, swapflag swp);
	  // schreibt die mittleren Sequenzen zurueck in den Baum
   
   /**********  erzeugen von Profilen  ****************************/
   void infer_six(int& vpro, int& xpro, substmodel& svx,
                  int upro, substmodel suv, int wpro, substmodel& svw,
	          int ypro, substmodel sxy, int zpro, substmodel& sxz);
   	  // ermittelt die Profile vpro und xpro fuer den Baum
      	  // W\   /Y
      	  //   V-X
      	  // U/   \Z
      	  // die Substitutionsmodelle unterscheiden sich nur durch ihre Zeiten
   void infer_five(int& vpro, int& xpro, substmodel& svx,
                   int upro, substmodel suv,
	           int ypro, substmodel sxy, int zpro, substmodel& sxz);
          // ermittelt die Profile vpro und xpro fuer den Baum
          //      /Y
          //   V-X
          // U/   \Z
          // die Substitutionsmodelle unterscheiden sich nur durch ihre Zeiten
   void infer_line(int& vpro, int& xpro, substmodel& svx,
                   int upro, substmodel suv, int zpro, substmodel& sxz);
          // ermittelt die Profile vpro und xpro fuer den Baum
          //      
          //   V-X
          // U/   \Z
          // die Substitutionsmodelle unterscheiden sich nur durch ihre Zeiten
   void infer_star(int& vpro, int& xpro, substmodel& svx,
	           int ypro, substmodel sxy, int zpro, substmodel& sxz);
          // ermittelt die Profile vpro und xpro fuer den Baum
          //      /Y
          //   V-X
          //      \Z
          // die Substitutionsmodelle unterscheiden sich nur durch ihre Zeiten
   void infer_three(int& vpro, int& xpro, substmodel& svx,
                    int zpro, substmodel& sxz);
          // ermittelt die Profile vpro und xpro fuer den Baum
          //      
          //   V-X
          //      \Z
          // die Substitutionsmodelle unterscheiden sich nur durch ihre Zeiten
   int infer_base(int profparent, substmodel& smparent,
                  int proflinks, substmodel& smlinks,
                  int profrechts, substmodel& smrechts);
   int infer_base(int profkind, substmodel& smkind);
   int infer_base(int proflinks, substmodel& smlinks,
                  int profrechts, substmodel& smrechts);

   /*************   Funktionen fuer das Viereralignment  ***********/
   void quartali(vector<quart_zus>& quart, baum *b, vector<unsigned int>& Kinder);
          // ermittelt das Alignment am Mercedesstern um Knoten b
   void quart_to_tri(vector<quart_zus>& quart,vector<triplet_zus>& tri);
          // ermittelt das Alignment der drei aeusseren Sequenzen
   quart_zus which_quartet_zus(vector<zustand>& paar_zus);
          // Welches Viereralignment wird durch drei paarweise Alignmnets beschrieben

   /*************   Funktionen fuer das Sechseralignment ***********/
   void hexali(baum *p, baum *q, int pidx, int qidx);
          // erzeugt ein neues Sechseralignment unter Beruecksichtigung
	  // der auszutauschenden Aeste
   swapflag which_pair_swaps(mercedes_stern& m1, mercedes_stern& m2,
                             mercedes_stern*& anc, mercedes_stern*& rec);
	  // Welchen Namen haben die Aeste, die ausgetauscht werden?
   void fillH(mercedes_stern& anc, mercedes_stern& rec, swapflag swp);
          // id.H werden die zu benutzenden Astlaengen uebergeben
   void zweimalvier(vector<quart_zus>& quart, mercedes_stern& m1, mercedes_stern& m2, swapflag swp);
          // Setzt die Alignments zweier benachbarter Mercedessterne zu einem
          // Alignment der vier aeusseren Sequenzen zusammen und beruecksichtigt
          // dabei, welche Aeste spaeter miteinander vertauscht werden.
   hex_zus two_quart_to_hex(mercedes_stern& m1, mercedes_stern& m2, vector<unsigned int>& idx);
          // Zentralknoten von m1 muss der Vater von dem von m2 sein!
          // Nimmt keine Ruecksicht auf die verschiedenen qY,qZ, hW und hY Zustaende!
          // Darf nicht fuer die Berechnung von Likelihoods verwendet werden!
          // aus X in m1.ali folgt W in m2.ali (und umgekehrt)
          // aus Z in m1.ali folgt X in m2.ali (und umgekehrt)
   quart_zus hex_to_quart(hex_zus hex);
          // gibt den Alignment zustand der vier aeusseren Knoten zurueck
          // Vertraegt sich nicht mit der Berechnung der Likelihood, da nicht
          // zwischen den verschiedenen qY und qZ Zustaenden unterschieden wird!
   quart_zus swap_quart_zus(quart_zus alt,swapflag swp);
          // Gibt neuen Zustand eine Viereralignments zurueck,
          // wenn die Topologie geaendert wurde (entspricht hier einer Umbenennung der Blaetter).
          // Die Wurzel (d.h. W) behaelt dabei immer ihren Namen. Deshalb fallen die Faelle
          // WggZ und XggY bzw. WggY und XggZ zusammen.
          // Vertraegt sich nicht mit der Berechnung der Likelihood, da nicht
          // zwischen den verschiedenen qY und qZ Zustaenden unterschieden wird!


   void optimaltime(baum *p);
	       
public :
   // Konstruktoren und Destruktor
   mult_alignment();
   mult_alignment(eingabe_parameter ep,int antyp);
   mult_alignment(const mult_alignment& mult1);
  ~mult_alignment();

   // Ausgabe
   void print_phylip(ostream& outalign);
   void print_screen(ostream& outalign);
   void print_params(ostream& pfstr);
   void print_tree(ostream& trstr);
   void bericht(ostream& logfile); //umfassender Zustandsbericht
   
   // Zustand
   double delta();   // gibt Mau-Newton-Parameter zurueck
   double likelihood();   // gibt lnL zurueck
   unsigned long int step();  // gibt zaehler zurueck
   bool burnin();                     // ist zaehler < coo.burnin
   bool inrange();                     // ist zaehler < coo.maxsteps
   bool report();                      // ist's Zeit fuer die Ausgabe
   unsigned int spez() const;  // OTU-Anzahl 

   // sampeln
   //void salterpearl();    // Sampeln des Baumes
   void toldWurz(); // neu Wurzeln (Likelihood wird neu berechnet)
   void maunewton();    // Sampeln des Baumes
   bool nearestneighbor(unsigned int wieviele);  // nearest-neighbour-interchange fuer told
   //void maunewton_fixedtopology();    // Sampeln neuer Astlaengen
   void sampparams();
   //void sampparams_version2();
   void optim_times();
   // Astlaengen optimieren
   void counterreset();
   // setzt zaehler um coo.reportafter zurueck
   void setcounter(unsigned long int c);
   // setzt zaehler auf c
};


class proposal {
public:
   vector<double> subpar;
   vector<double> idpar;
   double logcounter;
   double logdenominator;
   
   proposal(substmodel& sub, indel& id, keepfixed& kee);
  ~proposal();
   void propos_subpar(substmodel& sub, vector<bool>& subfix);
   void propos_idpar(indel& id, vector<bool>& indfix);
};

zustand get_zust(vector<int>& pAB, vector<int>& pBA, int a, int b);
          // Liefert den zu a und b assoziierten Alignment-Zustand
zustand iter_zust(vector<int>& AvsB, vector<int>& BvsA, vector<int>& indices);
          // Liefert den zu a und b assoziierten Alignment-Zustand
          // und setzt a und b einen Zustand weiter
bool patternequivalent(vector<char>& p1, vector<char>& p2);
          // sind an den Positionsmustern p1 und p2 die gleichen Taxa beteiligt?

#endif
