/****************************************************************************
 * This file is part of the program ALIFRITZ                                *
 *                                                                          *
 *    Copyright (C) 2001  Roland Fleissner                                  *
 *    e-mail: fleissne@uidaho.edu                            		    * 
 *                                                                          *
 *  This program is free software; you can redistribute it and/or modify    *
 *  it under the terms of the GNU General Public License as published by    *
 *  the Free Software Foundation; either version 2 of the License, or       *
 *  (at your option) any later version.                                     *
 *                                                                          *
 *  This program is distributed in the hope that it will be useful,         *
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of          *
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           *
 *  GNU General Public License for more details.                            *
 *                                                                          *
 *  You should have received a copy of the GNU General Public License       *
 *  along with this program; if not, write to the                           *
 *  Free Software Foundation, Inc., 59 Temple Place, Suite 330,             *
 *  Boston, MA  02111-1307  USA                                             *
 ****************************************************************************/

#ifndef PROFIL_H
#define PROFIL_H


/***********************************************************************************************/
/************     ein paar wichtige Funktionen             *************************************/
/***********************************************************************************************/

double addlog(double a, double b);// liefert log(exp(a)+exp(b))
double addlog(double a, double b, double c);// liefert log(exp(a)+exp(b)+exp(c))
double addlog(vector<double>& v);// liefert log(Summe_ueber_i von exp(v[i]))
double maximum(double a, double b);
double maximum(double a, double b, double c);
int maximum(int a, int b);
// distrib_pow ersetzt die Werte in distrib durch deren expo hoch itePot Potenz und renormiert
void distrib_pow(vector<double>& distrib, double expo, unsigned long int itePot); 


class alignment;

/***********************************************************************************************/
/************     class align_graph  und  Erben                        *************************/
/***********************************************************************************************/

class align_graph {
    friend ostream& operator<<(ostream& ostr, const align_graph& ag);
protected:
    int von_A, von_B, bis_A, bis_B;
    const alignment *algn;
    vector< vector<double> > A_gap_kante;
    vector< vector<double> > B_gap_kante;
    vector< vector<double> > paar_kante; 
private:
    void beschrifte()
	/* Beschriftet die Kanten des maxwk_align_graph fuer den angegebenen
	 * Bereich gemaess den DNA-Staengen und den Parametern von algn
	 */;
    void beschrifte(zustand firststate, zustand laststate)
	/* Beschriftet die Kanten des maxwk_align_graph fuer den angegebenen
	 * Bereich gemaess den DNA-Staengen und den Parametern von algn
	 */;
public:
    align_graph(const alignment& ali,
		int von_A,int von_B,int bis_A,int bis_B)
	/* Konstruiere einen Alignment-Graphen fuer das Teilalignment
	 * zwischen
	 * (von_A,von_B) und (bis_A,bis_B) in ali mit den aktuellen Parametern
	 * von algn
	 */;
    align_graph(const alignment& ali,
		int von_A,int von_B,int bis_A,int bis_B, zustand firststate, zustand laststate)
	/* Konstruiere einen Alignment-Graphen fuer das Teilalignment
	 * zwischen
	 * (von_A,von_B) und (bis_A,bis_B) in ali mit den aktuellen Parametern
	 * von algn
	 */;
    align_graph()
	/* Fuehrt zu Fehlermeldung
	 */;
    ~align_graph();

    zustand zufallsschritt(zustand s, int a, int b, const int PA, const int PB, double expo, unsigned long int itePot) const
	/* Gibt einen zufaelligen Zustand aus, der dem Schritt hin zu Knoten
	 * (a,b) entspricht, und zwar gemaess der bedingten Verteilung, dass
	 * es von (a,b) aus mit einem Schritt von Zustand s weitergeht.
	 */;
    zustand maxwk_schritt(zustand s, int a, int b, const int PA, const int PB) const
	/* Gibt einen wahrscheinlichsten Zustand aus, der dem Schritt hin zu Knoten
	 * (a,b) entspricht, und zwar gemaess der bedingten Verteilung, dass
	 * es von (a,b) aus mit einem Schritt von Zustand s weitergeht.
	 */;
};


ostream& operator<<(ostream& ostr, const align_graph& ag);


/***********************************************************************************************/
/************     class alignment                                      *************************/
/***********************************************************************************************/

class alignment {
    friend class align_graph;
private:
    const bioseq *pro_A, *pro_B;  // Sequenzen
    vector<int> pAB,pBA;            // Ausrichtung gegeneinander
    substmodel *sm;                   // Substitutionsmodell
    indel *idm;                     // Insertionsmodell
    double coolexp;                 // Exponent fuers Sampeln
    unsigned long int Schritt;           // Exponent des Exponenten
    vector<vector<double> > hmmtrans;  // nur fuer optimal_time von Bedeutung
    vector<vector<double> > basep;  // nur fuer optimal_time von Bedeutung
    
    void hirschberg(bool forEM);
        // speichersparendes Alignieren (bei forEM==TRUE werden hmmtrans und basep gefuellt
    void hirschberg(int von_A, int von_B, int bis_A, int bis_B, vector<zustand>& gesamtvec, zustand firststate, zustand laststate);
        // speichersparendes Alignieren
    void teilsample(int von_A, int von_B, int bis_A, int bis_B)
	/* sample ein neues Teilalignment von (von_A,von_B) 
	 * bis (bis_A,bis_B). Dabei bleiben die Zustaende der von
	 * Start- und Endpunkt erhalten.
	 */;
    double EMloglikl(double time,double scale);
        // Funktion, die im 'EM'-Alg. optimiert wird
    void twodimenmin(double tol,
	int active1, double min1, double& x1, double max1, double *err1,
	int active2, double min2, double& x2, double max2, double *err2);
    void onedimenmin(double, double, double, double&, double, double *, double *, double, bool);
    double brent(double, double, double, double, double *, double *, double, double, double, double, bool);
    void sample_paths();
public:
    alignment()
	/* Gibt eine Fehlermeldung
	 */;
    alignment(const bioseq &p_1,const bioseq &p_2, double time,
	      const substmodel &sub,const indel &ind,
	      double coolali, unsigned long int step)
	/* Initialisiert gemaess Substitutionsmodell sub, Indelmodell ind
	 * und unter Beruecksichtigung des SimAnneal-Schritts step
	 */;
    alignment(const bioseq &p_1,const bioseq &p_2, double time,
	      const substmodel &sub,const indel &ind)
	/* Konstruiert nicht wirklich ein Alignment, aber bereitet alles
	 * fuer den 'EM'-Alg. vor. Also Vorsicht!
	 */;
    alignment(const baum &b_1,const baum &b_2,
	      const substmodel &sub,const indel &ind,
	      double coolali, unsigned long int step)
	/* Initialisiert gemaess dem Alignment der Sequenzen an b_1
	 * und b_2, dem Substitutionsmodell sub, dem Indelmodell ind
	 * und unter Beruecksichtigung des SimAnneal-Schritts step
	 */;
    alignment(const alignment& align)
	/* Copy-Kostruktor
	 */;
    ~alignment();
    // const bioseq* get_prof_A() const {return pro_A;};
    // const bioseq* get_prof_B() const {return pro_B;};
    //const vector<int>& get_pAB() {return pAB;};
    //const vector<int>& get_pBA() {return pBA;};
    zustand get_zust(int a,int b) const
	/* Liefert den zu a und b assoziierten Alignment-Zustand
	 */; 
    zustand iter_zust(int* pa,int* pb) const
	/* Liefert den zu *pa und *pb assoziierten Alignment-Zustand und 
	 * setzt (*pa,*pb) einen Zustand weiter.
	 */;
    void sample()
	/* sample das Alignment
	 */;
    void teilsample(int von_A, int bis_A)
	/* sample ein neues Teilalignment von dna_A[von_A] 
	 * bis dna_A[bis_A].
	 */;
    void teilsample(int laenge)
	/* sample ein neues Teilalignment zu einem zufaelligen Abschnitt der 
	 * Laenge laenge. 
	 */;
    void zurueck(baum *b_1, baum *b_2)
        /* schreibt das ermittelte Alignment zurueck in den Baum
	 */;
    void optimal_time(double& time, double& scale);
        // optimiert die Zeit fuer 2 Sequenzen und gegebenes Modell
    void optimal_time(double& time);
        // optimiert die Zeit fuer 2 Sequenzen und gegebenes Modell
	// Gewichtung von Indel und Subst.prozess wird nicht geaendert
    double opt_for_giv();
        // optimiert die Zeit fuer ein paarweises Alignment
};

#endif
